@@ -1,5 +1,5 @@
 /*
-  Copyright 2007-2015 David Robillard <http://drobilla.net>
+  Copyright 2007-2016 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -49,21 +49,26 @@ typedef struct {
 	char* rel;  ///< Abstract path (relative path in state dir)
 } PathMap;
 
+typedef struct {
+	size_t    n;
+	Property* props;
+} PropertyArray;
+
 struct LilvStateImpl {
-	LilvNode*  plugin_uri;  ///< Plugin URI
-	LilvNode*  uri;         ///< State/preset URI
-	char*      dir;         ///< Save directory (if saved)
-	char*      file_dir;    ///< Directory for files created by plugin
-	char*      copy_dir;    ///< Directory for snapshots of external files
-	char*      link_dir;    ///< Directory for links to external files
-	char*      label;       ///< State/Preset label
-	ZixTree*   abs2rel;     ///< PathMap sorted by abs
-	ZixTree*   rel2abs;     ///< PathMap sorted by rel
-	Property*  props;       ///< State properties
-	PortValue* values;      ///< Port values
-	uint32_t   atom_Path;   ///< atom:Path URID
-	uint32_t   num_props;   ///< Number of state properties
-	uint32_t   num_values;  ///< Number of port values
+	LilvNode*     plugin_uri;   ///< Plugin URI
+	LilvNode*     uri;          ///< State/preset URI
+	char*         dir;          ///< Save directory (if saved)
+	char*         file_dir;     ///< Directory for files created by plugin
+	char*         copy_dir;     ///< Directory for snapshots of external files
+	char*         link_dir;     ///< Directory for links to external files
+	char*         label;        ///< State/Preset label
+	ZixTree*      abs2rel;      ///< PathMap sorted by abs
+	ZixTree*      rel2abs;      ///< PathMap sorted by rel
+	PropertyArray props;        ///< State properties
+	PropertyArray metadata;     ///< State metadata
+	PortValue*    values;       ///< Port values
+	uint32_t      atom_Path;    ///< atom:Path URID
+	uint32_t      n_values;     ///< Number of port values
 };
 
 static int
@@ -106,18 +111,18 @@ append_port_value(LilvState*  state,
                   uint32_t    size,
                   uint32_t    type)
 {
+	PortValue* pv = NULL;
 	if (value) {
 		state->values = (PortValue*)realloc(
-			state->values, (++state->num_values) * sizeof(PortValue));
-		PortValue* pv = &state->values[state->num_values - 1];
+			state->values, (++state->n_values) * sizeof(PortValue));
+		pv         = &state->values[state->n_values - 1];
 		pv->symbol = lilv_strdup(port_symbol);
 		pv->value  = malloc(size);
 		pv->size   = size;
 		pv->type   = type;
 		memcpy(pv->value, value, size);
-		return pv;
 	}
-	return NULL;
+	return pv;
 }
 
 static const char*
@@ -131,6 +136,32 @@ lilv_state_rel2abs(const LilvState* state, const char* path)
 	return path;
 }
 
+static void
+append_property(LilvState*     state,
+                PropertyArray* array,
+                uint32_t       key,
+                const void*    value,
+                size_t         size,
+                uint32_t       type,
+                uint32_t       flags)
+{
+	array->props = (Property*)realloc(
+		array->props, (++array->n) * sizeof(Property));
+
+	Property* const prop = &array->props[array->n - 1];
+	if ((flags & LV2_STATE_IS_POD) || type == state->atom_Path) {
+		prop->value = malloc(size);
+		memcpy(prop->value, value, size);
+	} else {
+		prop->value = (void*)value;
+	}
+
+	prop->size  = size;
+	prop->key   = key;
+	prop->type  = type;
+	prop->flags = flags;
+}
+
 static LV2_State_Status
 store_callback(LV2_State_Handle handle,
                uint32_t         key,
@@ -140,23 +171,7 @@ store_callback(LV2_State_Handle handle,
                uint32_t         flags)
 {
 	LilvState* const state = (LilvState*)handle;
-	state->props = (Property*)realloc(
-		state->props, (++state->num_props) * sizeof(Property));
-	Property* const prop = &state->props[state->num_props - 1];
-
-	if ((flags & LV2_STATE_IS_POD) || type == state->atom_Path) {
-		prop->value = malloc(size);
-		memcpy(prop->value, value, size);
-	} else {
-		LILV_WARN("Storing non-POD value\n");
-		prop->value = (void*)value;
-	}
-
-	prop->size  = size;
-	prop->key   = key;
-	prop->type  = type;
-	prop->flags = flags;
-
+	append_property(state, &state->props, key, value, size, type, flags);
 	return LV2_STATE_SUCCESS;
 }
 
@@ -170,7 +185,7 @@ retrieve_callback(LV2_State_Handle handle,
 	const LilvState* const state      = (LilvState*)handle;
 	const Property         search_key = { NULL, 0, key, 0, 0 };
 	const Property* const  prop       = (Property*)bsearch(
-		&search_key, state->props, state->num_props,
+		&search_key, state->props.props, state->props.n,
 		sizeof(Property), property_cmp);
 
 	if (prop) {
@@ -230,6 +245,7 @@ abstract_path(LV2_State_Map_Path_Handle handle,
 			char* copy  = lilv_get_latest_copy(real_path, cpath);
 			if (!copy || !lilv_file_equals(real_path, copy)) {
 				// No recent enough copy, make a new one
+				free(copy);
 				copy = lilv_find_free_path(cpath, lilv_path_exists, NULL);
 				const int st = lilv_copy_file(real_path, copy);
 				if (st) {
@@ -388,15 +404,15 @@ lilv_state_new_from_instance(const LilvPlugin*          plugin,
 			instance->lv2_handle, store_callback, state, flags, features);
 		if (st) {
 			LILV_ERRORF("Error saving plugin state: %s\n", state_strerror(st));
-			free(state->props);
-			state->props     = NULL;
-			state->num_props = 0;
+			free(state->props.props);
+			state->props.props = NULL;
+			state->props.n     = 0;
 		} else {
-			qsort(state->props, state->num_props, sizeof(Property), property_cmp);
+			qsort(state->props.props, state->props.n, sizeof(Property), property_cmp);
 		}
 	}
 
-	qsort(state->values, state->num_values, sizeof(PortValue), value_cmp);
+	qsort(state->values, state->n_values, sizeof(PortValue), value_cmp);
 
 	free(sfeatures);
 	return state;
@@ -407,7 +423,7 @@ lilv_state_emit_port_values(const LilvState*     state,
                             LilvSetPortValueFunc set_value,
                             void*                user_data)
 {
-	for (uint32_t i = 0; i < state->num_values; ++i) {
+	for (uint32_t i = 0; i < state->n_values; ++i) {
 		const PortValue* val = &state->values[i];
 		set_value(val->symbol, user_data, val->value, val->size, val->type);
 	}
@@ -430,19 +446,24 @@ lilv_state_restore(const LilvState*           state,
 		(LilvState*)state, abstract_path, absolute_path };
 	LV2_Feature map_feature = { LV2_STATE__mapPath, &map_path };
 
-	const LV2_Feature** sfeatures = add_features(features, &map_feature, NULL);
+	if (instance) {
+		const LV2_Descriptor* desc = instance->lv2_descriptor;
+		if (desc->extension_data) {
+			const LV2_State_Interface* iface = (const LV2_State_Interface*)
+				desc->extension_data(LV2_STATE__interface);
 
-	const LV2_Descriptor*      desc  = instance ? instance->lv2_descriptor : NULL;
-	const LV2_State_Interface* iface = (desc && desc->extension_data)
-		? (const LV2_State_Interface*)desc->extension_data(LV2_STATE__interface)
-		: NULL;
+			if (iface && iface->restore) {
+				const LV2_Feature** sfeatures = add_features(
+					features, &map_feature, NULL);
 
-	if (iface) {
-		iface->restore(instance->lv2_handle, retrieve_callback,
-		               (LV2_State_Handle)state, flags, sfeatures);
+				iface->restore(instance->lv2_handle, retrieve_callback,
+				               (LV2_State_Handle)state, flags, sfeatures);
+
+				free(sfeatures);
+			}
+		}
 	}
 
-	free(sfeatures);
 
 	if (set_value) {
 		lilv_state_emit_port_values(state, set_value, user_data);
@@ -494,7 +515,7 @@ new_state_from_model(LilvWorld*       world,
 		const SordNode* object = sord_iter_get_node(i, SORD_OBJECT);
 		const SordNode* graph  = sord_iter_get_node(i, SORD_GRAPH);
 		state->label = lilv_strdup((const char*)sord_node_get_string(object));
-		if (!state->dir) {
+		if (!state->dir && graph) {
 			state->dir = lilv_strdup((const char*)sord_node_get_string(graph));
 		}
 		sord_iter_free(i);
@@ -548,8 +569,9 @@ new_state_from_model(LilvWorld*       world,
 	if (state_node) {
 		SordIter* props = sord_search(model, state_node, 0, 0, 0);
 		FOREACH_MATCH(props) {
-			const SordNode* p = sord_iter_get_node(props, SORD_PREDICATE);
-			const SordNode* o = sord_iter_get_node(props, SORD_OBJECT);
+			const SordNode* p   = sord_iter_get_node(props, SORD_PREDICATE);
+			const SordNode* o   = sord_iter_get_node(props, SORD_OBJECT);
+			const char*     key = (const char*)sord_node_get_string(p);
 
 			chunk.len = 0;
 			lv2_atom_forge_set_sink(
@@ -560,7 +582,7 @@ new_state_from_model(LilvWorld*       world,
 			uint32_t        flags = LV2_STATE_IS_POD|LV2_STATE_IS_PORTABLE;
 			Property        prop  = { NULL, 0, 0, 0, flags };
 
-			prop.key   = map->map(map->handle, (const char*)sord_node_get_string(p));
+			prop.key   = map->map(map->handle, key);
 			prop.type  = atom->type;
 			prop.size  = atom->size;
 			prop.value = malloc(atom->size);
@@ -570,9 +592,9 @@ new_state_from_model(LilvWorld*       world,
 			}
 
 			if (prop.value) {
-				state->props = (Property*)realloc(
-					state->props, (++state->num_props) * sizeof(Property));
-				state->props[state->num_props - 1] = prop;
+				state->props.props = (Property*)realloc(
+					state->props.props, (++state->props.n) * sizeof(Property));
+				state->props.props[state->props.n - 1] = prop;
 			}
 		}
 		sord_iter_free(props);
@@ -583,8 +605,12 @@ new_state_from_model(LilvWorld*       world,
 	free((void*)chunk.buf);
 	sratom_free(sratom);
 
-	qsort(state->props, state->num_props, sizeof(Property), property_cmp);
-	qsort(state->values, state->num_values, sizeof(PortValue), value_cmp);
+	if (state->props.props) {
+		qsort(state->props.props, state->props.n, sizeof(Property), property_cmp);
+	}
+	if (state->values) {
+		qsort(state->values, state->n_values, sizeof(PortValue), value_cmp);
+	}
 
 	return state;
 }
@@ -700,7 +726,9 @@ ttl_writer(SerdSink sink, void* stream, const SerdNode* base, SerdEnv** new_env)
 
 	SerdWriter* writer = serd_writer_new(
 		SERD_TURTLE,
-		(SerdStyle)(SERD_STYLE_RESOLVED|SERD_STYLE_ABBREVIATED|SERD_STYLE_CURIED),
+		(SerdStyle)(SERD_STYLE_RESOLVED |
+		            SERD_STYLE_ABBREVIATED|
+		            SERD_STYLE_CURIED),
 		env,
 		&base_uri,
 		sink,
@@ -856,6 +884,37 @@ link_exists(const char* path, void* data)
 	return !matches;
 }
 
+static void
+write_property_array(const LilvState*     state,
+                     const PropertyArray* array,
+                     Sratom*              sratom,
+                     uint32_t             flags,
+                     const SerdNode*      subject,
+                     LV2_URID_Unmap*      unmap,
+                     const char*          dir)
+{
+	for (uint32_t i = 0; i < array->n; ++i) {
+		Property*   prop = &array->props[i];
+		const char* key  = unmap->unmap(unmap->handle, prop->key);
+
+		const SerdNode p = serd_node_from_string(SERD_URI, USTR(key));
+		if (prop->type == state->atom_Path && !dir) {
+			const char* path     = (const char*)prop->value;
+			const char* abs_path = lilv_state_rel2abs(state, path);
+			LILV_WARNF("Writing absolute path %s\n", abs_path);
+			sratom_write(sratom, unmap, flags,
+			             subject, &p, prop->type,
+			             strlen(abs_path) + 1, abs_path);
+		} else if (prop->flags & LV2_STATE_IS_POD ||
+		           prop->type == state->atom_Path) {
+			sratom_write(sratom, unmap, flags,
+			             subject, &p, prop->type, prop->size, prop->value);
+		} else {
+			LILV_WARNF("Lost non-POD property <%s> on save\n", key);
+		}
+	}
+}
+
 static int
 lilv_state_write(LilvWorld*       world,
                  LV2_URID_Map*    map,
@@ -902,11 +961,14 @@ lilv_state_write(LilvWorld*       world,
 	                (SerdEndSink)serd_writer_end_anon,
 	                writer);
 
-	// Write port values as pretty numbers
-	sratom_set_pretty_numbers(sratom, true);
+	// Write metadata
+	sratom_set_pretty_numbers(sratom, false);  // Use precise types
+	write_property_array(state, &state->metadata, sratom, 0,
+	                     &subject, unmap, dir);
 
 	// Write port values
-	for (uint32_t i = 0; i < state->num_values; ++i) {
+	sratom_set_pretty_numbers(sratom, true);  // Use pretty numbers
+	for (uint32_t i = 0; i < state->n_values; ++i) {
 		PortValue* const value = &state->values[i];
 
 		const SerdNode port = serd_node_from_string(
@@ -931,35 +993,19 @@ lilv_state_write(LilvWorld*       world,
 		serd_writer_end_anon(writer, &port);
 	}
 
-	// Write property values with precise types
-	sratom_set_pretty_numbers(sratom, false);
-
 	// Write properties
-	const SerdNode state_node = serd_node_from_string(SERD_BLANK,
-	                                                  USTR("2state"));
-	if (state->num_props > 0) {
+	const SerdNode body = serd_node_from_string(SERD_BLANK, USTR("body"));
+	if (state->props.n > 0) {
 		p = serd_node_from_string(SERD_URI, USTR(LV2_STATE__state));
 		serd_writer_write_statement(writer, SERD_ANON_O_BEGIN, NULL,
-		                            &subject, &p, &state_node, NULL, NULL);
+		                            &subject, &p, &body, NULL, NULL);
 	}
-	for (uint32_t i = 0; i < state->num_props; ++i) {
-		Property*   prop = &state->props[i];
-		const char* key  = unmap->unmap(unmap->handle, prop->key);
+	sratom_set_pretty_numbers(sratom, false);  // Use precise types
+	write_property_array(state, &state->props, sratom, SERD_ANON_CONT,
+	                     &body, unmap, dir);
 
-		p = serd_node_from_string(SERD_URI, USTR(key));
-		if (prop->type == state->atom_Path && !dir) {
-			const char* path     = (const char*)prop->value;
-			const char* abs_path = lilv_state_rel2abs(state, path);
-			sratom_write(sratom, unmap, SERD_ANON_CONT,
-			             &state_node, &p, prop->type,
-			             strlen(abs_path) + 1, abs_path);
-		} else {
-			sratom_write(sratom, unmap, SERD_ANON_CONT,
-			             &state_node, &p, prop->type, prop->size, prop->value);
-		}
-	}
-	if (state->num_props > 0) {
-		serd_writer_end_anon(writer, &state_node);
+	if (state->props.n > 0) {
+		serd_writer_end_anon(writer, &body);
 	}
 
 	sratom_free(sratom);
@@ -1116,12 +1162,12 @@ lilv_state_delete(LilvWorld*       world,
 		model, state->uri->node, world->uris.rdfs_seeAlso, NULL, NULL);
 	if (file) {
 		// Remove state file
-		char* file_path = lilv_file_uri_parse(
+		char* path = lilv_file_uri_parse(
 			(const char*)sord_node_get_string(file), NULL);
-		if (unlink(file_path)) {
-			LILV_ERRORF("Failed to remove %s (%s)\n", file_path, strerror(errno));
+		if (unlink(path)) {
+			LILV_ERRORF("Failed to remove %s (%s)\n", path, strerror(errno));
 		}
-		lilv_free(file_path);
+		lilv_free(path);
 	}
 
 	// Remove any existing manifest entries for this state
@@ -1158,14 +1204,26 @@ lilv_state_delete(LilvWorld*       world,
 	return 0;
 }
 
+static void
+free_property_array(LilvState* state, PropertyArray* array)
+{
+	for (uint32_t i = 0; i < array->n; ++i) {
+		Property* prop = &array->props[i];
+		if ((prop->flags & LV2_STATE_IS_POD) ||
+		    prop->type == state->atom_Path) {
+			free(prop->value);
+		}
+	}
+	free(array->props);
+}
+
 LILV_API void
 lilv_state_free(LilvState* state)
 {
 	if (state) {
-		for (uint32_t i = 0; i < state->num_props; ++i) {
-			free(state->props[i].value);
-		}
-		for (uint32_t i = 0; i < state->num_values; ++i) {
+		free_property_array(state, &state->props);
+		free_property_array(state, &state->metadata);
+		for (uint32_t i = 0; i < state->n_values; ++i) {
 			free(state->values[i].value);
 			free(state->values[i].symbol);
 		}
@@ -1173,7 +1231,6 @@ lilv_state_free(LilvState* state)
 		lilv_node_free(state->uri);
 		zix_tree_free(state->abs2rel);
 		zix_tree_free(state->rel2abs);
-		free(state->props);
 		free(state->values);
 		free(state->label);
 		free(state->dir);
@@ -1191,12 +1248,12 @@ lilv_state_equals(const LilvState* a, const LilvState* b)
 	    || (a->label && !b->label)
 	    || (b->label && !a->label)
 	    || (a->label && b->label && strcmp(a->label, b->label))
-	    || a->num_props != b->num_props
-	    || a->num_values != b->num_values) {
+	    || a->props.n != b->props.n
+	    || a->n_values != b->n_values) {
 		return false;
 	}
 
-	for (uint32_t i = 0; i < a->num_values; ++i) {
+	for (uint32_t i = 0; i < a->n_values; ++i) {
 		PortValue* const av = &a->values[i];
 		PortValue* const bv = &b->values[i];
 		if (av->size != bv->size || av->type != bv->type
@@ -1206,9 +1263,9 @@ lilv_state_equals(const LilvState* a, const LilvState* b)
 		}
 	}
 
-	for (uint32_t i = 0; i < a->num_props; ++i) {
-		Property* const ap = &a->props[i];
-		Property* const bp = &b->props[i];
+	for (uint32_t i = 0; i < a->props.n; ++i) {
+		Property* const ap = &a->props.props[i];
+		Property* const bp = &b->props.props[i];
 		if (ap->key != bp->key
 		    || ap->type != bp->type
 		    || ap->flags != bp->flags) {
@@ -1230,7 +1287,7 @@ lilv_state_equals(const LilvState* a, const LilvState* b)
 LILV_API unsigned
 lilv_state_get_num_properties(const LilvState* state)
 {
-	return state->num_props;
+	return state->props.n;
 }
 
 LILV_API const LilvNode*
@@ -1258,3 +1315,15 @@ lilv_state_set_label(LilvState* state, const char* label)
 	state->label = (char*)realloc(state->label, len + 1);
 	memcpy(state->label, label, len + 1);
 }
+
+LILV_API int
+lilv_state_set_metadata(LilvState*  state,
+                        uint32_t    key,
+                        const void* value,
+                        size_t      size,
+                        uint32_t    type,
+                        uint32_t    flags)
+{
+	append_property(state, &state->metadata, key, value, size, type, flags);
+	return LV2_STATE_SUCCESS;
+}
