@@ -1,5 +1,5 @@
 /*
-  Copyright 2007-2015 David Robillard <http://drobilla.net>
+  Copyright 2007-2016 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -27,25 +27,27 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <stddef.h>
 
 #ifdef _WIN32
+#ifndef _WIN32_WINNT
 #    define _WIN32_WINNT 0x0600  /* for CreateSymbolicLink */
+#endif
 #    include <windows.h>
 #    include <direct.h>
 #    include <io.h>
 #    define F_OK 0
 #    define mkdir(path, flags) _mkdir(path)
-#    if (defined(_MSC_VER) && (_MSC_VER < 1500))
 /** Implement 'CreateSymbolicLink()' for MSVC 8 or earlier */
 BOOLEAN WINAPI
 CreateSymbolicLink(LPCTSTR linkpath, LPCTSTR targetpath, DWORD flags)
 {
 	typedef BOOLEAN (WINAPI* PFUNC)(LPCTSTR, LPCTSTR, DWORD);
 
-	PFUNC pfn = (PFUNC)GetProcAddress(GetModuleHandle(TEXT("kernel32.dll")), "CreateSymbolicLinkA");
+	PFUNC pfn = (PFUNC)GetProcAddress(GetModuleHandle(TEXT("kernel32.dll")),
+	                                  "CreateSymbolicLinkA");
 	return pfn ? pfn(linkpath, targetpath, flags) : 0;
 }
-#    endif /* _MSC_VER < 1500 */
 #else
 #    include <dirent.h>
 #    include <limits.h>
@@ -62,7 +64,7 @@ CreateSymbolicLink(LPCTSTR linkpath, LPCTSTR targetpath, DWORD flags)
 #endif
 
 #ifndef PAGE_SIZE
-#   define PAGE_SIZE 4096
+#    define PAGE_SIZE 4096
 #endif
 
 void
@@ -297,6 +299,7 @@ lilv_copy_file(const char* src, const char* dst)
 
 	FILE* out = fopen(dst, "w");
 	if (!out) {
+		fclose(in);
 		return errno;
 	}
 
@@ -371,26 +374,8 @@ lilv_path_join(const char* a, const char* b)
 	return path;
 }
 
-static void
-lilv_size_mtime(const char* path, off_t* size, time_t* time)
-{
-	struct stat buf;
-	if (stat(path, &buf)) {
-		LILV_ERRORF("stat(%s) (%s)\n", path, strerror(errno));
-		return;
-	}
-
-	if (size) {
-		*size = buf.st_size;
-	}
-	if (time) {
-		*time = buf.st_mtime;
-	}
-}
-
 typedef struct {
 	char*  pattern;
-	off_t  orig_size;
 	time_t time;
 	char*  latest;
 } Latest;
@@ -398,16 +383,18 @@ typedef struct {
 static void
 update_latest(const char* path, const char* name, void* data)
 {
-	Latest* latest     = (Latest*)data;
-	char*   entry_path = lilv_path_join(path, name);
+	Latest*  latest     = (Latest*)data;
+	char*    entry_path = lilv_path_join(path, name);
 	unsigned num;
 	if (sscanf(entry_path, latest->pattern, &num) == 1) {
-		off_t  entry_size = 0;
-		time_t entry_time = 0;
-		lilv_size_mtime(entry_path, &entry_size, &entry_time);
-		if (entry_size == latest->orig_size && entry_time >= latest->time) {
-			free(latest->latest);
-			latest->latest = entry_path;
+		struct stat st;
+		if (!stat(entry_path, &st)) {
+			if (st.st_mtime >= latest->time) {
+				free(latest->latest);
+				latest->latest = entry_path;
+			}
+		} else {
+			LILV_ERRORF("stat(%s) (%s)\n", path, strerror(errno));
 		}
 	}
 	if (entry_path != latest->latest) {
@@ -420,8 +407,14 @@ char*
 lilv_get_latest_copy(const char* path, const char* copy_path)
 {
 	char*  copy_dir = lilv_dirname(copy_path);
-	Latest latest   = { lilv_strjoin(copy_path, "%u", NULL), 0, 0, NULL };
-	lilv_size_mtime(path, &latest.orig_size, &latest.time);
+	Latest latest   = { lilv_strjoin(copy_path, ".%u", NULL), 0, NULL };
+
+	struct stat st;
+	if (!stat(path, &st)) {
+		latest.time = st.st_mtime;
+	} else {
+		LILV_ERRORF("stat(%s) (%s)\n", path, strerror(errno));
+	}
 
 	lilv_dir_for_each(copy_dir, &latest, update_latest);
 
@@ -613,19 +606,16 @@ lilv_file_equals(const char* a_path, const char* b_path)
 	FILE*       b_file = NULL;
 	char* const a_real = lilv_realpath(a_path);
 	char* const b_real = lilv_realpath(b_path);
-	if (!a_real || !b_real) {
-		match = false;  // Missing file matches nothing
-	} else if (!strcmp(a_real, b_real)) {
+	if (!strcmp(a_real, b_real)) {
 		match = true;  // Real paths match
 	} else if (lilv_file_size(a_path) != lilv_file_size(b_path)) {
 		match = false;  // Sizes differ
-	} else if (!(a_file = fopen(a_real, "rb"))) {
-		match = false;  // Missing file matches nothing
-	} else if (!(b_file = fopen(b_real, "rb"))) {
+	} else if (!(a_file = fopen(a_real, "rb")) ||
+	           !(b_file = fopen(b_real, "rb"))) {
 		match = false;  // Missing file matches nothing
 	} else {
-		match = true;
 		// TODO: Improve performance by reading chunks
+		match = true;
 		while (!feof(a_file) && !feof(b_file)) {
 			if (fgetc(a_file) != fgetc(b_file)) {
 				match = false;
